# rFeedback() handles the \shinyRFeedback environment for innerCode()


rFeedback = function(text, inputId, options, lineNumOffset, globalInfoEnv) {
  if (inputId=="" || make.names(inputId)!=inputId) {
    stop("@ ", lineNumOffset, 
         " first argument of \\begin{shinyRFeedback} must be a valid R name (inputId)")
  }
  
  options = parseCommand(paste0("shinyRFeedback(second_curly_brace_set){", options, "}"), 
                         lineNumOffset)$options
  options = parseOptions("shinyRFeedback", options, 
                         required=NULL, requiredTypes=NULL,
                         optional=c("nrows","ncols", "prompt","label"), 
                         optionalTypes=c("n","n","c","c"),
                         lineNumOffset=lineNumOffset,
                         defaults=list(nrows=1, ncols=30, prompt=">",
                                       label="Feedback:"))
  
  lineCnt = options$nrows
  if (lineCnt<1 || lineCnt>30 || lineCnt!=round(lineCnt)) 
    stop("@ ", lineNumOffset, " 'lines' option of \\shinyRFeedback must be from 1 to 30")
  
  # Find block markers in recipe           
  recipe = str_trim(text[2:(length(text)-1)])
  emptyLoc = grep("^EMPTY:", recipe)
  if (length(emptyLoc) > 1) {
    stop("@ ", paste(lineNumOffset+emptyLoc, collapse=", "), 
         " \\shinyRFeedback environments can have only one 'EMPTY:'")
  }
  nefariousLoc = grep("^NEFARIOUS:", recipe)
  if (length(nefariousLoc) > 1) {
    stop("@ ", paste(lineNumOffset+nefariousLoc, collapse=", "), 
         " \\shinyRFeedback environments can have only one 'NEFARIOUS:'")
  }
  calculateLoc = grep("^CALCULATE:", recipe)
  if (length(calculateLoc) > 1) {
    stop("@ ", paste(lineNumOffset+calculateLoc, collapse=", "), 
         " \\shinyRFeedback environments can have only one 'CALCULATE:'")
  }
  errorAndLocs = grep("^ERRORAND[[:space:]]*[(].+[)][[:space:]]*:$", recipe)
  errorLoc = grep("^ERROR:", recipe)
  if (length(errorLoc) > 1) {
    stop("@ ", paste(lineNumOffset+errorLoc, collapse=", "), 
         " \\shinyRFeedback environments can have only one 'ERROR:'")
  }
  inHasLocs = grep("^INPUTHAS[[:space:]]*[(].+[)][[:space:]]*:$", recipe)
  inLacksLocs = grep("^INPUTLACKS[[:space:]]*[(].+[)][[:space:]]*:$", recipe)
  outHasLocs = grep("^OUTPUTHAS[[:space:]]*[(].+[)][[:space:]]*:$", recipe)
  outLacksLocs = grep("^OUTPUTLACKS[[:space:]]*[(].+[)][[:space:]]*:$", recipe)
  envEmptyLocs = grep("^ENVEMPTY[[:space:]]*:$", recipe)
  envNotEmptyLocs = grep("^ENVNOTEMPTY[[:space:]]*:$", recipe)
  envHasLocs = grep("^ENVHAS[[:space:]]*[(].+[)][[:space:]]*:$", recipe)
  envLacksLocs = grep("^ENVLACKS[[:space:]]*[(].+[)][[:space:]]*:$", recipe)
  envClassIsLocs = grep("^ENV[\\[][a-zA-Z._0-9]+]CLASSIS[[:space:]]*[(].+):$", recipe)
  envClassIsntLocs = grep("^ENV[\\[][a-zA-Z._0-9]+]CLASSISNT[[:space:]]*[(].+):$", recipe)
  valueMeetsLocs = grep("^VALUE[\\[][a-zA-Z._0-9]+]MEETS[[:space:]]*[(].+):$", recipe)
  defaultLoc = grep("^DEFAULT:", recipe)
  
  # Some checks
  if (length(defaultLoc) > 1) {
    stop("@ ", paste(lineNumOffset+defaultLoc, collapse=", "), 
         " \\shinyRFeedback environments can have only one 'DEFAULT:'")
  }
  if (length(errorLoc)>0 && length(errorAndLocs)>0 && any(errorAndLocs>errorLoc)) {
    warning("@ ", paste(lineNumOffset+c(errorLoc, errorAndLocs[errorAndLocs>errorLoc]), collapse=", "), 
            " \\shinyRFeedback reordered 'ERROR:' to follow all 'ERRORAND:'s")
  }
  
  # report on misspelled block markers
  allStartLocs = c(emptyLoc, nefariousLoc, calculateLoc, errorAndLocs, errorLoc, 
                   inHasLocs, inLacksLocs, outHasLocs, outLacksLocs, 
                   envEmptyLocs, envNotEmptyLocs, envHasLocs, envLacksLocs, 
                   envClassIsLocs, envClassIsntLocs, valueMeetsLocs, defaultLoc)
  extra = setdiff(grep("^[a-zA-Z]+.*:$", recipe), allStartLocs)
  if (length(extra) > 0) {
    stop("@ ", lineNumOffset+extra,
         " unrecognized or malformed \\shinyRFeedback block marker(s): ", recipe[extra[1]])
  }
  
  # Get EMPTY, CALCULATE, NEFARIOUS, and DEFAULT feedback blocks or my default values
  borderLocs = c(allStartLocs, length(recipe)+1)
  # get text from 'recipe' starting after 'start' or return default (and remove "#" comments)
  readText = function(start, default) {
    if (length(start)==0) return(default)
    end = min(borderLocs[borderLocs>start]) - 1
    if (start==end) return(default)
    text = recipe[(start+1):end]
    text = text[grep("$[[:space:]]*#", text, invert=TRUE)]
    if (length(text)==0) return(default)
    return(text)
  }
  emptyText = readText(emptyLoc, '"(no input)"')
  nefariousText = readText(nefariousLoc, '"Please do not try to use such dangerous commands."')
  calculateText = readText(calculateLoc, NA)
  defaultText = readText(defaultLoc, '"Your code produced no errors, but the output is not as expected."')
  errorText = readText(errorLoc, '"Your code produced an error."')

  
  # Set User Interface code
  ########## To get textEnter to respond only at Enter, go to
  ########## C:\Users\hseltman\Documents\R\win-library\3.2\shiny\www\shared
  ########## and erase "$(a).on("keyup.textInputBinding input.textInputBinding",function(a){b(!0)}),"
  if (lineCnt == 1) {
    feedInput = paste0('textInput(inputId="readRCode', inputId, '", label="',
                       options$prompt, '", width="', options$ncols, 'em"),')
  } else {
    feedInput = paste0('textareaInput(inputId="readRCode', inputId, '",',
                       'nrows=', lineCnt, ', ncols=', options$ncols, '),')
  }
  uiCode = c(feedInput,
             'HTML(\'<div style="padding:1ex"></div>\'),',
             paste0('HTML("<p><strong>', options$label, '</strong></p>"),'),
             paste0('verbatimTextOutput(outputId="rCodeFeedbackOut', inputId, '"),'),
             'HTML(\'<div style="padding:1ex"></div>\'),',
             #paste0('checkboxInput(inputId="getRDiagnose', inputId,'", label="Diagnose errors", value=FALSE),'),
             #paste0('conditionalPanel("input.getRDiagnose', inputId, ' == true",'),
             #paste0('                 verbatimTextOutput(outputId="rCodeDiagnoseOut', inputId, '")'),
             #'),',
             'HTML(\'<div style="padding:1ex"></div>\'),')
  
  # create text for a giant if () { foo } else if { bar } else if { spam } ...
  ifBlock = function(textCondition, text, noelse=FALSE, noif=FALSE, indent=2) {
    indent = paste(rep(" ",indent), collapse="")
    indent2 = paste0(indent, indent)
    rtn = paste0(indent, ifelse(noelse, "", "} else "))
    if (noif) {
      rtn = paste0(rtn, "} else {")
    } else {
      rtn = paste0(rtn, "if (", textCondition, ") {")
    }
    quoted = grep("^[\"'].*[\"']$", text)
    if (any(quoted)) {
      text[quoted] = paste0('cat(', substring(text[quoted], 1, nchar(text[quoted])-1), '\\n")')
    }
    return(c(rtn, paste0(indent2, text, sep="")))
  }

  # Set initial server code
  serverCode = c(paste0('readRCodeRtn', inputId, ' = reactive({'),
                 paste0('  VARS = list2env(list())'),
                 paste0('  rtn = runR(input$readRCode', inputId, ', VARS)'),
                 '  return(c(rtn, list(VARS=VARS)))',
                 '})',
                 paste0('output$rCodeFeedbackOut', inputId, '=renderPrint({'),
                 paste0('  INPUT = isolate(input$readRCode', inputId, ')'),
                 paste0('  OUTPUT = readRCodeRtn', inputId, '()$output'),
                 paste0('  NEFARIOUS = readRCodeRtn', inputId, '()$nefarious'),
                 paste0('  EMPTY = is.null(readRCodeRtn', inputId, '()$output) && ',
                            'is.null(readRCodeRtn', inputId, '()$error)'),
                 paste0('  ERROR = ! is.null(readRCodeRtn', inputId, '()$error)'),
                 paste0('  ERRORCODE = readRCodeRtn', inputId, '()$errorCode'),
                 paste0('  ERRORTEXT = readRCodeRtn', inputId, '()$error'),
                 paste0('  VARS = readRCodeRtn', inputId, '()$VARS'),
                 paste0('  VARNAMES = ls(envir=VARS)'))


  # Setup 'caseOrder" with all info for each case (except EMPTY, NEFARIOUS, CALCULATE, and DEFAULT)
  caseOrder = NULL
  if (length(errorAndLocs>0)) caseOrder = rbind(caseOrder, data.frame(start=errorAndLocs, type="errorAnd"))
  if (length(inHasLocs>0)) caseOrder = rbind(caseOrder, data.frame(start=inHasLocs, type="inHas"))
  if (length(inLacksLocs>0)) caseOrder = rbind(caseOrder, data.frame(start=inLacksLocs, type="inLacks"))
  if (length(outHasLocs>0)) caseOrder = rbind(caseOrder, data.frame(start=outHasLocs, type="outHas"))
  if (length(outLacksLocs>0)) caseOrder = rbind(caseOrder, data.frame(start=outLacksLocs, type="outLacks"))
  if (length(envEmptyLocs>0)) caseOrder = rbind(caseOrder, data.frame(start=envEmptyLocs, type="envEmpty"))
  if (length(envNotEmptyLocs>0)) caseOrder = rbind(caseOrder, data.frame(start=envNotEmptyLocs, type="envNotEmpty"))
  if (length(envHasLocs>0)) caseOrder = rbind(caseOrder, data.frame(start=envHasLocs, type="envHas"))
  if (length(envLacksLocs>0)) caseOrder = rbind(caseOrder, data.frame(start=envLacksLocs, type="envLacks"))
  if (length(envClassIsLocs>0)) caseOrder = rbind(caseOrder, data.frame(start=envClassIsLocs, type="envClassIs"))
  if (length(envClassIsntLocs>0)) caseOrder = rbind(caseOrder, data.frame(start=envClassIsntLocs, type="envClassIsnt"))
  if (length(valueMeetsLocs>0)) caseOrder = rbind(caseOrder, data.frame(start=valueMeetsLocs, type="valueMeets"))
  
  getInside = function(x, delim=c("parentheses","brackets")) {
    delim = match.arg(delim)
    if (delim=="parentheses") {
      re = "[(].+)"
    } else {
      re = "[\\[].+]"
    }
    m = regexpr(re, x)
    return(substring(x, m+1, m+attr(m,"match.length")-2))
  }
  
  # Store variable name for feedback blocks with variables        
  caseOrder$var = ""
  hasVarLocs = c(envClassIsLocs, envClassIsntLocs, valueMeetsLocs)
  if (length(hasVarLocs) > 0) {
    vars = sapply(hasVarLocs, function(line) getInside(recipe[line], delim="brackets"))
    caseOrder[match(hasVarLocs, caseOrder$start), "var"] = vars
  }
  
  # Store and syntax check condition expressions for feedback blocks with conditions
  caseOrder$condition = ""
  hasCondLocs = c(errorAndLocs, inHasLocs, inLacksLocs,
                  outHasLocs, outLacksLocs, envHasLocs, envLacksLocs, envClassIsLocs, 
                  envClassIsntLocs, valueMeetsLocs)
  if (length(hasCondLocs) > 0) {
    conds = sapply(hasCondLocs, function(line) getInside(recipe[line], delim="paren"))
    #condBad = sapply(conds, function(x) {is(try(parse(text=x), silent=TRUE), "try-error")})
    Sel = match(hasCondLocs, caseOrder$start)
    #if (any(condBad)) {
    #  stop("invalid feedback block conditions(s) for RFeedback '", inputId, "': ",
    #       paste(paste0(recipe[caseOrder$start[Sel[condBad]]], " (@", 
    #                    lineNumOffset+caseOrder$start[Sel[condBad]],
    #                    ")"), collapse=", "))
    #}
    caseOrder[Sel, "condition"] = conds
  }
  
  # Store ending rows for feedback blocks
  borderLocs = c(allStartLocs, length(recipe)+1)
  caseOrder$end = sapply(caseOrder$start,
                         function(x) min(borderLocs[borderLocs>x])-1)
  # Warn of empty blocks
  Sel = caseOrder$start == caseOrder$end
  if (any(Sel)) {
    warning("empty block(s) for RFeedback '", inputId, "': ",
            paste(paste0(caseOrder[Sel,"type"], " (@", 
                         lineNumOffset+caseOrder[Sel,"start"], ")"), collapse=", "))
    caseOrder = caseOrder[!Sel,]
  }
  
  # Sort in user's order
  if (nrow(caseOrder)>0) caseOrder = caseOrder[order(caseOrder$start),]

  # Start giant (simulated) "switch" (EMPTY and NEFARIOUS)
  serverCode = c(serverCode,
                 ifBlock("EMPTY", emptyText, noelse=TRUE),
                 ifBlock("NEFARIOUS", nefariousText))
  
    
  # Continue producing server code for "switch" (ERROR and ERRORAND)
  Sel = caseOrder$type == "errorAnd"
  if (any(Sel)) {
    errorAnd = caseOrder[Sel, ]
    caseOrder = caseOrder[!Sel, ]
    for (i in 1:nrow(errorAnd)) {
      serverCode = c(serverCode, 
                     ifBlock(paste("ERROR &&", errorAnd[i, "condition"]), 
                                   recipe[(errorAnd[i, "start"]+1):(errorAnd[i, "end"])]))
    }
  }
  serverCode = c(serverCode,
                 ifBlock("ERROR", errorText))
  
  # Continue producing server code for "switch" (now just ENVHAS and ENVLACKS)
  blockText = lapply(1:nrow(caseOrder),
                     function(i) recipe[(caseOrder[i, "start"]+1):(caseOrder[i, "end"])])
  for (i in 1:nrow(caseOrder)) {
    if (caseOrder[i, "type"] == "inHas") {
      serverCode = c(serverCode,
                     ifBlock(paste0('length(grep("', caseOrder[i, "condition"], '", INPUT)) != 0'), 
                             blockText[[i]])
      )
    } else if (caseOrder[i, "type"] == "inLacks") {
      serverCode = c(serverCode,
                     ifBlock(paste0('length(grep("', caseOrder[i, "condition"], '", INPUT)) == 0'), 
                             blockText[[i]])
      )
    }
    if (caseOrder[i, "type"] == "outHas") {
      serverCode = c(serverCode,
                     ifBlock(paste0('length(grep("', caseOrder[i, "condition"], '", OUTPUT)) != 0'), 
                             blockText[[i]])
      )
    } else if (caseOrder[i, "type"] == "outLacks") {
      serverCode = c(serverCode,
                     ifBlock(paste0('length(grep("', caseOrder[i, "condition"], '", OUTPUT)) == 0'), 
                             blockText[[i]])
      )
    }
    if (caseOrder[i, "type"] == "envHas") {
      serverCode = c(serverCode,
                     ifBlock(paste0('"', caseOrder[i, "condition"], '" %in% VARNAMES'), 
                             blockText[[i]])
      )
    } else if (caseOrder[i, "type"] == "envLacks") {
      serverCode = c(serverCode,
                     ifBlock(paste0('! "', caseOrder[i, "condition"], '" %in% VARNAMES'), 
                             blockText[[i]])
      )
    }
    if (caseOrder[i, "type"] == "envEmpty") {
      serverCode = c(serverCode,
                     ifBlock(paste0('length(VARNAMES) == 0'), 
                             blockText[[i]])
      )
    } else if (caseOrder[i, "type"] == "envNotEmpty") {
      serverCode = c(serverCode,
                     ifBlock(paste0('length(VARNAMES) != 0'), 
                             blockText[[i]])
      )
    }
    if (caseOrder[i, "type"] == "envClassIs") {
      serverCode = c(serverCode,
                     ifBlock(paste0('("', caseOrder[i, "var"], '" %in% VARNAMES) && ', 
                                    'is(get("', caseOrder[i, "var"], '", envir=VARS), "',
                                         caseOrder[i, "condition"], '")'), 
                             blockText[[i]])
      )
    }
    if (caseOrder[i, "type"] == "envClassIsnt") {
      serverCode = c(serverCode,
                     ifBlock(paste0('("', caseOrder[i, "var"], '" %in% VARNAMES) && ', 
                                    '! is(get("',caseOrder[i, "var"], '", envir=VARS), "',
                                    caseOrder[i, "condition"], '")'), 
                             blockText[[i]])
      )
    }
    if (caseOrder[i, "type"] == "valueMeets") {
      condition0 = paste0('("', caseOrder[i, "var"], '" %in% VARNAMES)')
      condition1 = gsub("VAR", 
                        paste0('(get("',caseOrder[i, "var"], '", envir=VARS))'),
                        caseOrder[i, "condition"], fixed=TRUE)
      safeCondition = paste0(condition0, ' && !is(try(', condition1, ', silent=TRUE), "try-error") && ', condition1)
      serverCode = c(serverCode,
                     ifBlock(safeCondition, blockText[[i]])
      )
    }
  }
  
  
  # Complete server code "switch" (DEFAULT)
  serverCode = c(serverCode,
                 ifBlock("", defaultText, noelse=TRUE, noif=TRUE),
                 "  }", "})")
  

  return(list(uiCode=uiCode, serverCode=serverCode))
}